package Locale::CLDR::LanguageMatching;
# This file auto generated from Data\common\supplemental\languageInfo.xml
#	on Sat  6 Jul  6:44:11 am GMT

use strict;
use warnings;
use version;

our $VERSION = version->declare('v0.45.0');

use v5.10.1;
use mro 'c3';
use utf8;
use if $^V ge v5.12.0, feature => 'unicode_strings';
use Types::Standard qw( Str Int HashRef ArrayRef CodeRef RegexpRef );
use Moo::Role;

has 'paradigm_locales' => (
	is			=> 'ro',
	isa			=> ArrayRef,
	init_arg	=> undef,
	default		=> sub {
		return [ qw( en en_GB es es_419 pt_BR pt_PT ) ],
	},
);

has 'language_match' => (
	is			=> 'ro',
	isa			=> ArrayRef,
	init_arg	=> undef,
	default		=> sub {
		return [
			{
				d_language	=> 'nb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '1',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'no',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'bs',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'bs',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ssy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'aa',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gsw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'de',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'de',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'da',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '8',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'no',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'da',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '8',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'nb',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ab',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ach',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'af',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'nl',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ak',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'am',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ay',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'az',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bal',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ur',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'be',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bem',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'br',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ca',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ceb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fil',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'chr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ckb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'co',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'crs',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'cs',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sk',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'cy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ee',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'eo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'eu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'da',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'nl',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ga',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gaa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gl',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ha',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'haw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ht',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ia',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ig',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'is',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'jv',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'id',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ka',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'km',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kri',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ku',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'tr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ky',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'la',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'it',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ln',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'loz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lua',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mai',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mfe',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mi',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ml',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ms',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'id',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'my',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ne',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'nb',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'no',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nso',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ny',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nyn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'oc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'om',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'or',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pcm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ps',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'rm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'de',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'rn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'rw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'si',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'so',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'st',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'su',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'id',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ta',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'te',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ti',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tlh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'to',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tum',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ug',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'uk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ur',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'uz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'wo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'fr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'xh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'yi',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'yo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'za',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'zu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '30',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aao',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'abh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'abv',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'acm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'acq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'acw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'acx',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'acy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'adf',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aeb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aec',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'afb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'apc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'apd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'arq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ars',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ary',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'arz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'auz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'avl',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ayh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ayl',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ayn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ayp',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bbz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pga',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'shu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ssh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'azb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'az',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'vro',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'et',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ffm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fub',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fue',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fuf',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fuh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fui',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fuq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'fuv',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ff',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gnw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'gn',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gui',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'gn',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gun',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'gn',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nhd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'gn',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ikt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'iu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'enb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'eyo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'niq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'oki',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pko',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sgc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tec',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tuy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kln',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gom',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kok',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gkp',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kpe',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ida',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lkb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lko',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lks',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lri',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lrm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lsm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lto',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lts',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lwg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nle',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nyd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'rag',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'luy',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ltg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'lv',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bhr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bjq',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bmm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bzc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'msh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'skg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tdx',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tkg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'txy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'xmv',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'xmw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mg',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mvf',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mn',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bjn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'btj',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bve',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'bvu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'coa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'dup',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hji',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'id',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'jak',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'jax',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kvb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kvr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'kxd',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lce',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lcf',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'liw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'max',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'meo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mfa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mfb',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'min',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mqg',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'msi',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mui',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'orn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ors',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pel',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pse',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'tmw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'urk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'vkk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'vkt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'xmm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'zlm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'zmi',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ms',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'dty',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ne',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gax',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'om',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hae',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'om',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'orc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'om',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'spv',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'or',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pbt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ps',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pst',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ps',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qub',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qud',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'quf',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qug',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'quh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'quk',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qul',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qup',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qur',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qus',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'quw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qux',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'quy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qva',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qve',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvi',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvj',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvl',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvm',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvp',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvs',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qvz',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qwa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qwc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qwh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qws',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxa',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxl',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxp',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxr',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'qxw',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'qu',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sdc',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sc',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sdn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sc',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'sro',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sc',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aae',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sq',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aat',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sq',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aln',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sq',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'aii',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'syr',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'uzs',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'uz',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'yih',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'yi',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'cdo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'cjy',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'cpx',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'czh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'czo',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'gan',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hak',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'hsn',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'lzh',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'mnp',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'nan',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'wuu',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'yue',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> '*',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '80',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> '*',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'am',
				d_region	=> '*',
				d_script	=> 'Ethi',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'az',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> 'Cyrl',
			},
			{
				d_language	=> 'bn',
				d_region	=> '*',
				d_script	=> 'Beng',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'bo',
				d_region	=> '*',
				d_script	=> 'Tibt',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hans',
			},
			{
				d_language	=> 'hy',
				d_region	=> '*',
				d_script	=> 'Armn',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> 'Cyrl',
			},
			{
				d_language	=> 'ka',
				d_region	=> '*',
				d_script	=> 'Geor',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'km',
				d_region	=> '*',
				d_script	=> 'Khmr',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'kn',
				d_region	=> '*',
				d_script	=> 'Knda',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'lo',
				d_region	=> '*',
				d_script	=> 'Laoo',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'ml',
				d_region	=> '*',
				d_script	=> 'Mlym',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'my',
				d_region	=> '*',
				d_script	=> 'Mymr',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'ne',
				d_region	=> '*',
				d_script	=> 'Deva',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'or',
				d_region	=> '*',
				d_script	=> 'Orya',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'pa',
				d_region	=> '*',
				d_script	=> 'Guru',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'ps',
				d_region	=> '*',
				d_script	=> 'Arab',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'sd',
				d_region	=> '*',
				d_script	=> 'Arab',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'si',
				d_region	=> '*',
				d_script	=> 'Sinh',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'ta',
				d_region	=> '*',
				d_script	=> 'Taml',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'te',
				d_region	=> '*',
				d_script	=> 'Telu',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'ti',
				d_region	=> '*',
				d_script	=> 'Ethi',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'tk',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> 'Cyrl',
			},
			{
				d_language	=> 'ur',
				d_region	=> '*',
				d_script	=> 'Arab',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'uz',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ru',
				s_region	=> '*',
				s_script	=> 'Cyrl',
			},
			{
				d_language	=> 'yi',
				d_region	=> '*',
				d_script	=> 'Hebr',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> 'Latn',
			},
			{
				d_language	=> 'sr',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'sr',
				s_region	=> '*',
				s_script	=> 'Cyrl',
			},
			{
				d_language	=> 'za',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '10',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hans',
			},
			{
				d_language	=> 'zh',
				d_region	=> '*',
				d_script	=> 'Hani',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hans',
			},
			{
				d_language	=> 'zh',
				d_region	=> '*',
				d_script	=> 'Hani',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'ar',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> 'Arab',
			},
			{
				d_language	=> 'bn',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'bn',
				s_region	=> '*',
				s_script	=> 'Beng',
			},
			{
				d_language	=> 'gu',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'gu',
				s_region	=> '*',
				s_script	=> 'Gujr',
			},
			{
				d_language	=> 'hi',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'hi',
				s_region	=> '*',
				s_script	=> 'Deva',
			},
			{
				d_language	=> 'kn',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'kn',
				s_region	=> '*',
				s_script	=> 'Knda',
			},
			{
				d_language	=> 'ml',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ml',
				s_region	=> '*',
				s_script	=> 'Mlym',
			},
			{
				d_language	=> 'mr',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'mr',
				s_region	=> '*',
				s_script	=> 'Deva',
			},
			{
				d_language	=> 'ta',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ta',
				s_region	=> '*',
				s_script	=> 'Taml',
			},
			{
				d_language	=> 'te',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'te',
				s_region	=> '*',
				s_script	=> 'Telu',
			},
			{
				d_language	=> 'zh',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '20',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hans',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Latn',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Jpan',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Hani',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Jpan',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Hira',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Jpan',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Kana',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Jpan',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Hrkt',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Jpan',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Hira',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Hrkt',
			},
			{
				d_language	=> 'ja',
				d_region	=> '*',
				d_script	=> 'Kana',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ja',
				s_region	=> '*',
				s_script	=> 'Hrkt',
			},
			{
				d_language	=> 'ko',
				d_region	=> '*',
				d_script	=> 'Hani',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ko',
				s_region	=> '*',
				s_script	=> 'Kore',
			},
			{
				d_language	=> 'ko',
				d_region	=> '*',
				d_script	=> 'Hang',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ko',
				s_region	=> '*',
				s_script	=> 'Kore',
			},
			{
				d_language	=> 'ko',
				d_region	=> '*',
				d_script	=> 'Jamo',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ko',
				s_region	=> '*',
				s_script	=> 'Kore',
			},
			{
				d_language	=> 'ko',
				d_region	=> '*',
				d_script	=> 'Jamo',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ko',
				s_region	=> '*',
				s_script	=> 'Hang',
			},
			{
				d_language	=> '*',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '50',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> '*',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'DZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'TN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'LY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'MR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MA',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'DZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'TN',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'LY',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'MR',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> 'EH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'ar',
				s_region	=> 'EH',
				s_script	=> '*',
			},
			{
				d_language	=> 'ar',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'ar',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '3',
				not_dr	=> '1',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> 'GB',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'AS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'GU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'MP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PH',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'UM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'AS',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'GU',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'MP',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PH',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'UM',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'en',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'en',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'en',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'es',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'es',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'es',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'US',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BZ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'NI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PA',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BB',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CU',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CW',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'DO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GD',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GP',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'HT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'JM',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KN',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'KY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'LC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MQ',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'MS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SX',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'TT',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VG',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VI',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'AR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'BV',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CL',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'CO',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'EC',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'FK',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GF',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GS',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'GY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'PY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'SR',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'UY',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'US',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BZ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'NI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PA',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BB',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CU',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CW',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'DO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GD',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GP',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'HT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'JM',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KN',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'KY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'LC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MQ',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'MS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SX',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'TT',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VG',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VI',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'AR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'BV',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CL',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'CO',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'EC',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'FK',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GF',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GS',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'GY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'PY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'SR',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'UY',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> 'VE',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'pt',
				s_region	=> 'VE',
				s_script	=> '*',
			},
			{
				d_language	=> 'pt',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'pt',
				s_region	=> '*',
				s_script	=> '*',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'HK',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> 'HK',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'HK',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> 'MO',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'MO',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> 'HK',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'MO',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> 'MO',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'HK',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'zh',
				s_region	=> 'HK',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'HK',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'zh',
				s_region	=> 'MO',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'MO',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'zh',
				s_region	=> 'HK',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> 'MO',
				d_script	=> 'Hant',
				distance	=> '4',
				not_dr	=> '1',
				not_sr	=> '1',
				s_language	=> 'zh',
				s_region	=> 'MO',
				s_script	=> 'Hant',
			},
			{
				d_language	=> 'zh',
				d_region	=> '*',
				d_script	=> 'Hant',
				distance	=> '5',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> 'zh',
				s_region	=> '*',
				s_script	=> 'Hant',
			},
			{
				d_language	=> '*',
				d_region	=> '*',
				d_script	=> '*',
				distance	=> '4',
				not_dr	=> '0',
				not_sr	=> '0',
				s_language	=> '*',
				s_region	=> '*',
				s_script	=> '*',
			},
		];
	},
);

no Moo::Role;

1;

# vim: tabstop=4
